@@ -9,7 +9,7 @@ from OpenSSL import crypto
 from django_afip import exceptions
 
 
-def create_embeded_pkcs7_signature(data: bytes, cert: bytes, key: bytes):
+def create_embeded_pkcs7_signature(data: bytes, cert: bytes, key: bytes) -> bytes:
     """Creates an embedded ("nodetached") PKCS7 signature.
 
     This is equivalent to the output of::
@@ -60,7 +60,3 @@ def create_csr(key_file, organization_name, common_name, serial_number, file_):
     req.sign(key, "md5")
 
     file_.write(crypto.dump_certificate_request(crypto.FILETYPE_PEM, req))
-
-
-def parse_certificate(file_):
-    return crypto.load_certificate(crypto.FILETYPE_PEM, file_)
